import sys
from io import StringIO
from unittest.mock import patch

from ocdskit.cli.__main__ import main
from tests import path


def test_command(monkeypatch):
    with patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'schema-report', '--min-occurrences', '2',
                                          path('test-schema.json')])
        main()

    assert actual.getvalue() == '''codelist,openCodelist
a.csv,False/True
b.csv,False
c.csv,False
d.csv,False

 2: {'codelist': 'a.csv', 'openCodelist': True, 'type': ['string', 'null']}
'''


def test_command_no_codelists(monkeypatch):
    with patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'schema-report', '--min-occurrences', '2', '--no-codelists',
                                          path('test-schema.json')])
        main()

    assert 'codelist,openCodelist' not in actual.getvalue()
    assert ':' in actual.getvalue()


def test_command_no_definitions(monkeypatch):
    with patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'schema-report', '--min-occurrences', '2', '--no-definitions',
                                          path('test-schema.json')])
        main()

    assert 'codelist,openCodelist' in actual.getvalue()
    assert ':' not in actual.getvalue()


def test_command_min_occurrences(monkeypatch):
    with patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'schema-report', '--min-occurrences', '1', '--no-codelists',
                                          path('test-schema.json')])
        main()

    assert 'codelist,openCodelist' not in actual.getvalue()
    assert '1:' in actual.getvalue()



from collections import defaultdict, OrderedDict

from ocdsextensionregistry import ProfileBuilder
from ocdsmerge.merge import merge, merge_versioned, get_tags, get_release_schema_url

from ocdskit.exceptions import InconsistentVersionError
from ocdskit.util import get_ocds_minor_version


def package_releases(releases, uri='', publisher=None, published_date='', extensions=None):
    """
    Wraps releases in a release package.

    :param list releases: a list of releases
    :param str uri: the release package's ``uri``
    :param dict publisher: the release package's ``publisher``
    :param str published_date: the release package's ``publishedDate``
    :param list extensions: the release package's ``extensions``
    """
    if publisher is None:
        publisher = OrderedDict()
    if extensions is None:
        extensions = []

    output = OrderedDict([
        ('uri', uri),
        ('publisher', publisher),
        ('publishedDate', published_date),
        ('version', '1.1'),
        ('extensions', extensions),
        ('releases', releases),
    ])

    return output


def combine_record_packages(packages, uri='', publisher=None, published_date=''):
    """
    Collects the packages and records from the record packages into one record package.

    :param list packages: a list of record packages
    :param str uri: the record package's ``uri``
    :param dict publisher: the record package's ``publisher``
    :param str published_date: the record package's ``publishedDate``
    """
    if publisher is None:
        publisher = OrderedDict()

    output = OrderedDict([
        ('uri', uri),
        ('publisher', publisher),
        ('publishedDate', published_date),
        ('license', None),
        ('publicationPolicy', None),
        ('version', None),
        ('extensions', OrderedDict()),
        ('packages', []),
        ('records', []),
    ])

    for package in packages:
        _update_package_metadata(output, package, publisher)

        output['records'].extend(package['records'])

        if 'packages' in package:
            output['packages'].extend(package['packages'])

    if not output['packages']:
        del output['packages']

    _set_extensions_metadata(output)
    _remove_empty_optional_metadata(output)

    return output


def combine_release_packages(packages, uri='', publisher=None, published_date=''):
    """
    Collects the releases from the release packages into one release package.

    :param list packages: a list of release packages
    :param str uri: the release package's ``uri``
    :param dict publisher: the release package's ``publisher``
    :param str published_date: the release package's ``publishedDate``
    """
    if publisher is None:
        publisher = OrderedDict()

    output = OrderedDict([
        ('uri', uri),
        ('publisher', publisher),
        ('publishedDate', published_date),
        ('license', None),
        ('publicationPolicy', None),
        ('version', None),
        ('extensions', OrderedDict()),
        ('releases', []),
    ])

    for package in packages:
        _update_package_metadata(output, package, publisher)

        output['releases'].extend(package['releases'])

    _set_extensions_metadata(output)
    _remove_empty_optional_metadata(output)

    return output


def compile_release_packages(packages, uri='', publisher=None, published_date='', schema=None,
                             return_versioned_release=False, return_package=False, use_linked_releases=False):
    """
    Merges releases by OCID and yields compiled releases.

    If ``return_versioned_release`` is ``True``, yields the versioned release. If ``return_package`` is ``True``, wraps
    the compiled releases (and versioned releases if ``return_versioned_release`` is ``True``) in a record package.

    If ``return_package`` is set and ``publisher`` isn't set, the output record package will have the same publisher as
    the last input release package.

    :param list packages: a list of release packages
    :param str uri: if ``return_package`` is ``True``, the record package's ``uri``
    :param dict publisher: if ``return_package`` is ``True``, the record package's ``publisher``
    :param str published_date: if ``return_package`` is ``True``, the record package's ``publishedDate``
    :param dict schema: the URL or path of the release schema to use
    :param bool return_package: wrap the compiled releases in a record package
    :param bool use_linked_releases: if ``return_package`` is ``True``, use linked releases instead of full releases
    :param bool return_versioned_release: if ``return_package`` is ``True``, include versioned releases in the record
        package; otherwise, yield versioned releases instead of compiled releases
    """
    if return_package:
        output = OrderedDict([
            ('uri', uri),
            ('publisher', publisher),
            ('publishedDate', published_date),
            ('license', None),
            ('publicationPolicy', None),
            ('version', None),
            ('extensions', OrderedDict()),
            ('packages', []),
            ('records', []),
        ])
    # To avoid duplicating code, we track extensions in the same place even if ``return_package`` is false.
    else:
        output = {
            'extensions': OrderedDict(),
        }

    version = None
    releases_by_ocid = defaultdict(list)
    linked_releases = []

    for i, package in enumerate(packages):
        if not version:
            version = get_ocds_minor_version(package)
        else:
            v = get_ocds_minor_version(package)
            if v != version:
                raise InconsistentVersionError('item {}: version error: this package uses version {}, but earlier '
                                               'packages used version {}'.format(i, v, version), version, v)

        if not schema:
            prefix = version.replace('.', '__') + '__'
            tag = next(tag for tag in reversed(get_tags()) if tag.startswith(prefix))
            schema = get_release_schema_url(tag)

        for release in package['releases']:
            releases_by_ocid[release['ocid']].append(release)

            if return_package and use_linked_releases:
                linked_releases.append(OrderedDict([
                    ('url', package['uri'] + '#' + release['id']),
                    ('date', release['date']),
                    ('tag', release['tag']),
                ]))

        if return_package:
            _update_package_metadata(output, package, publisher)

            output['packages'].append(package['uri'])
        else:
            _update_extensions_metadata(output, package)

    if output['extensions']:
        builder = ProfileBuilder(tag, list(output['extensions']))
        schema = builder.patched_release_schema()

    if return_package:
        for ocid, releases in releases_by_ocid.items():
            record = OrderedDict([
                ('ocid', ocid),
                ('releases', []),
                ('compiledRelease', merge(releases, schema)),
            ])

            if use_linked_releases:
                record['releases'] = linked_releases
            else:
                record['releases'] = releases

            if return_versioned_release:
                record['versionedRelease'] = merge_versioned(releases, schema)

            output['records'].append(record)

        _set_extensions_metadata(output)
        _remove_empty_optional_metadata(output)

        yield output
    else:
        for releases in releases_by_ocid.values():
            if return_versioned_release:
                merge_method = merge_versioned
            else:
                merge_method = merge

            merged_release = merge_method(releases, schema)

            yield merged_release


def _update_package_metadata(output, package, publisher):
    _update_extensions_metadata(output, package)

    if not publisher and 'publisher' in package:
        output['publisher'] = package['publisher']

    for field in ('license', 'publicationPolicy', 'version'):
        if field in package:
            output[field] = package[field]


def _update_extensions_metadata(output, package):
    if 'extensions' in package:
        # Python has no OrderedSet, so we use OrderedDict to keep extensions in order without duplication.
        output['extensions'].update(OrderedDict.fromkeys(package['extensions']))


def _set_extensions_metadata(output):
    if output['extensions']:
        output['extensions'] = list(output['extensions'])
    else:
        del output['extensions']


def _remove_empty_optional_metadata(output):
    for field in ('license', 'publicationPolicy', 'version'):
        if output[field] is None:
            del output[field]



from .base import OCDSCommand
from ocdskit import upgrade
from ocdskit.exceptions import CommandError


class Command(OCDSCommand):
    name = 'upgrade'
    help = 'upgrades packages and releases from an old version of OCDS to a new version'

    def add_arguments(self):
        self.add_argument('versions', help='the colon-separated old and new versions')

    def handle(self):
        versions = self.args.versions

        version_from, version_to = versions.split(':')
        if version_from < version_to:
            direction = 'up'
        else:
            direction = 'down'

        try:
            upgrade_method = getattr(upgrade, 'upgrade_{}'.format(versions.replace('.', '').replace(':', '_')))
        except AttributeError:
            raise CommandError('{}grade from {} is not supported'.format(direction, versions.replace(':', ' to ')))

        for data in self.items():
            data = upgrade_method(data)
            self.print(data)



import sys
from io import BytesIO, StringIO, TextIOWrapper
from unittest.mock import patch

from ocdskit.cli.__main__ import main
from tests import read


def test_command(monkeypatch):
    stdin = read('release_minimal-1.json', 'rb') + read('release_minimal-2.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'package-releases', '--uri', 'http://example.com',
                                          '--published-date', '9999-01-01T00:00:00Z', '--publisher-name', ''])
        main()

    assert actual.getvalue() == read('release-package_minimal-1-2.json')


def test_command_extensions(monkeypatch):
    stdin = read('release_minimal-1.json', 'rb') + read('release_minimal-2.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'package-releases', '--uri', 'http://example.com',
                                          '--published-date', '9999-01-01T00:00:00Z', '--publisher-name', '',
                                          'http://example.com/a/extension.json',
                                          'http://example.com/b/extension.json'])
        main()

    assert actual.getvalue() == read('release-package_minimal-1-2-extensions.json')


def test_command_root_path_array(monkeypatch):
    stdin = read('realdata/record-package-1.json', 'rb') + read('realdata/record-package-2.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'package-releases', '--root-path', 'records.item.releases'])
        main()

    assert actual.getvalue() == read('realdata/release-package_record-package.json')


def test_command_root_path_item(monkeypatch):
    stdin = read('realdata/record-package-1.json', 'rb') + read('realdata/record-package-2.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'package-releases', '--root-path', 'records.item.releases.item'])
        main()

    assert actual.getvalue() == read('realdata/release-package_record-package.json')



import json
import sys
from io import BytesIO, StringIO, TextIOWrapper
from unittest.mock import patch

from ocdskit.cli.__main__ import main
from tests import read


def test_command(monkeypatch):
    stdin = read('release-package_minimal.json', 'rb') + \
            read('release-package_maximal.json', 'rb') + \
            read('release-package_extensions.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'combine-release-packages'])
        main()

    assert actual.getvalue() == read('combine-release-packages_minimal-maximal-extensions.json')


def test_command_no_extensions(monkeypatch):
    stdin = read('release-package_minimal.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'combine-release-packages'])
        main()

    assert actual.getvalue() == read('combine-release-packages_minimal.json')


def test_command_uri_published_date(monkeypatch):
    stdin = read('release-package_minimal.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'combine-release-packages', '--uri', 'http://example.com/x.json',
                                          '--published-date', '2010-01-01T00:00:00Z'])
        main()

    package = json.loads(actual.getvalue())
    assert package['uri'] == 'http://example.com/x.json'
    assert package['publishedDate'] == '2010-01-01T00:00:00Z'


def test_command_publisher(monkeypatch):
    stdin = read('release-package_minimal.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'combine-release-packages', '--publisher-name', 'Acme Inc.'])
        main()

    package = json.loads(actual.getvalue())
    assert package['publisher']['name'] == 'Acme Inc.'
